import inspect
import warnings


def raise_warn_call(*warn_cats):
    """Decorator to transform warnings into exception during a function call

    Usage:
        # This will transform all warnings into exception
        @raise_warn_call
        def fun(*args, **kwargs):
            pass

        # This will transform the specified warnings into exception
        @raise_warn_call(SyntaxWarning, RuntimeWarning)
        def fun(*args, **kwargs):
            pass
    """
    # If only a single element which is not a warning is passed, filter all warnings
    is_fun = (len(warn_cats) == 1) and (
        (not inspect.isclass(warn_cats[0])) or (not issubclass(warn_cats[0], Warning))
    )

    # A function is passed
    if is_fun:
        fun = warn_cats[0]

        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                print(*args, **kwargs)
                out = fun(*args, **kwargs)
            return out

        return wrapper

    # Warnings are passed
    def deco(fun):
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                for cat in warn_cats:
                    warnings.filterwarnings("error", category=cat)
                return fun(*args, **kwargs)

        return wrapper

    return deco



from typing import Any, Callable

X1 = Any
X2 = Any
Y1 = Any
Y2 = Any


def interpretation(fun: Callable) -> Callable[[Callable[..., Y1]], Callable[..., Y1]]:
    """
    Decorator for composition on the right
    Take as argument fun, outputs a decorator transformin function g(x, ...) into g(fun(x), ...)
    Further arguments passed when constructing the decorator are passed to fun.
    """

    def deco(gfun: Callable[..., Y1], *fargs, **fkwargs) -> Callable[..., Y1]:
        def wrapper(x, *args, **kwargs):
            return gfun(fun(x, *fargs, **fkwargs), *args, **kwargs)

        return wrapper

    return deco


def post_modif(
    fun: Callable[[Y1], Y2]
) -> Callable[[Callable[..., Y1]], Callable[..., Y2]]:
    """
    Decorator for composition on the left.
    Take as argument fun, outputs a decorator transformin function g into fun(g( ...))
    Further arguments passed when constructing the decorator are passed to fun.
    """

    def deco(gfun: Callable[..., Y1], *fargs, **fkwargs) -> Callable[..., Y2]:
        def wrapper(*args, **kwargs):
            return fun(gfun(*args, **kwargs), *fargs, **fkwargs)

        return wrapper

    return deco



from functools import partial
from typing import Any, Callable, Iterable

from multiprocess import Pool  # pylint:disable=E0611

Input = Any
Output = Any


def par_eval(
    fun: Callable[..., Output], xs: Iterable[Input], parallel: bool, *args, **kwargs
) -> list[Output]:
    """
    Evaluation of a function on a list of values. If parallel is True,
    computations are parallelized using multiprocess.Pool . Else list
    comprehension is used.

    Further arguments and keyword arguments are passed to fun.
    """
    if parallel:
        loc_fun = partial(fun, *args, **kwargs)
        with Pool() as pool:  # pylint: disable=E1102
            out = pool.map(loc_fun, xs)
    else:
        out = [fun(x, *args, **kwargs) for x in xs]
    return out



from typing import Any

import numpy as np
import pandas as pd

from apicutils.sets import are_set_equal


def get_last_valid_index(df: pd.DataFrame) -> tuple[int, Any]:
    """From dataframe df with potential nan value,
    return the highest row location count such that
    df.iloc[0:count] has no Nan, as well as the index of the value
    df.index[count-1] (None if count = 0)"""
    df_nan = np.isnan(df.to_numpy())
    index_nan = np.any(df_nan, axis=1)

    # First index is not valid
    if index_nan[0]:
        return 0, None

    # All indexes are valid
    if np.all(~index_nan):
        return len(index_nan), df.index[-1]

    # At least a good index (0) and at least a bad index
    count = np.argmin(~index_nan)
    return count, df.index[count - 1]


# EQUALITY CHECKS
def frames_is_close(
    df_1: pd.DataFrame, df_2: pd.DataFrame, rtol: float = 1e-5, atol: float = 1e-8
) -> bool:
    """
    Checks if two dataframes are almost equal.
    Dataframes are assumed to contain only numeric values. Checks are performed using
    np.allclose function with equal_nan=True. Checks for equality of index and columns
    are performed (shuffles of columns/index will not break equality).

    Args:
        df_1, df_2: the two dataframes to be compared to one another
        kwargs: rtol, atol, passed to numpy allclose function

    TODO:
        Adjust to englobe cases where dataframes contain non
        numerical columns, or have inadequate dtype (object)
    """

    if not are_set_equal(df_1.index, df_2.index):
        return False
    if not are_set_equal(df_1.columns, df_2.columns):
        return False

    return np.allclose(
        df_1, df_2.loc[df_1.index][df_1.columns], equal_nan=True, rtol=rtol, atol=atol
    )


def frames_coincide(
    df_1: pd.DataFrame, df_2: pd.DataFrame, rtol: float = 1e-5, atol: float = 1e-8
) -> bool:
    """Check that df_2 coincide with df_1 on df_1 index
    Args:
        df_1, df_2: Two dataframes
    Returns:
        True if df_1's index is contained in df_2 and if df_1 and df_2 coincide
        on this index, False if not
    """
    idx_min = df_1.index
    if set(idx_min).difference(df_2.index):
        # index of df_1 is not contained in df_2
        return False
    return frames_is_close(df_1, df_2.loc[idx_min], atol=atol, rtol=rtol)


def frames_equal(df_1: pd.DataFrame, df_2: pd.DataFrame) -> bool:
    """Check that two dataframes are equal, up to a reordering of index/columns"""
    if not are_set_equal(df_1.index, df_2.index):
        return False
    if not are_set_equal(df_1.columns, df_2.columns):
        return False

    return ((df_1 - df_2) == 0.0).all().all()



"""
miscealleanous shared between different packages

Functions:
    blab: wrap up for silent
    timedrun: TimeOut option for function call decorator
    interpretation: right composition decorator
    post_modif: left composition decorator
    safe_call: Evaluation of any function without failure (returns None if Exception occured)
    par_eval: Parallelisation switch for function evaluation
    num_der: numerical differentiation
    vectorize: function vectorization
"""

from apicutils.composition import interpretation, post_modif
from apicutils.dataframes import get_last_valid_index
from apicutils.matrix_inversion import safe_inverse_ps_matrix
from apicutils.num_der import num_der
from apicutils.parallel import par_eval
from apicutils.prints import blab, tprint
from apicutils.raise_warn_call import raise_warn_call
from apicutils.safe_call import SafeCallWarning, safe_call
from apicutils.sets import are_set_equal, check_set_equal, set_diff_msg
from apicutils.shape import ShapeError, check_shape, get_pre_shape, prod
from apicutils.timed_run import timedrun, timeout
from apicutils.vectorize import vectorize



import numpy as np


def safe_inverse_ps_matrix(matrix: np.ndarray, eps: float = 10**-6) -> np.ndarray:
    """Compute the inverse of symmetric positive definite matrix.
    Correct np.linalg.inv implementation when matrix has bad conditionning number.

    Inversion of eigenvalues < eps will be heavily perturbed.
    This function is still work in progress
    """

    return np.linalg.inv(0.5 * (matrix + matrix.T) + eps * np.eye(len(matrix)))



def set_diff_msg(iter_1, iter_2) -> str:
    err_msgs = []
    extra_1 = set(iter_1).difference(iter_2)
    if extra_1:
        err_msgs.append(f"Index 1 contains extra keys {extra_1}")
    extra_2 = set(iter_2).difference(iter_1)
    if extra_2:
        err_msgs.append(f"Index 2 contains extra keys {extra_2}")

    err_msg = "\nFurthermore:".join(err_msgs)

    return err_msg


def are_set_equal(iter_1, iter_2) -> bool:
    err_msg = set_diff_msg(iter_1, iter_2)
    if err_msg:
        return False
    return True


def check_set_equal(iter_1, iter_2):
    """Check if two set like object have identical keys. If not, raise a ValueError"""
    err_msg = set_diff_msg(iter_1, iter_2)

    if err_msg:
        raise ValueError(err_msg)



from typing import Callable

import numpy as np

from apicutils.parallel import par_eval
from apicutils.shape import prod


def vectorize(
    fun: Callable,
    input_shape: tuple[int, ...],
    convert_input: bool = True,
    parallel: bool = True,
) -> Callable:
    """For a function fun which takes as input np.ndarray of shape input_shape and outputs
    arrays of shape output_shape, outputs the vectorized function which takes as input np.ndarray
    of shape (pre_shape, input_shape) and outputs np.ndarrat of shape (pre_shape, output_shape)
    """
    d = len(input_shape)

    def new_fun(xs) -> np.ndarray:
        if convert_input:
            xs = np.array(xs)
        pre_shape = xs.shape[:-d]
        xs.reshape(
            (
                (
                    prod(
                        pre_shape,
                    ),
                )
                + input_shape
            )
        )
        out = np.array(par_eval(fun, xs, parallel=parallel))
        out.reshape(pre_shape + out.shape[1:])
        return out

    return new_fun



import numpy as np


class ShapeError(ValueError):
    """Exception class when array shape is not as expected"""


def check_shape(xs: np.ndarray, shape_exp: tuple[int, ...]) -> None:
    """Check if xs has a given shape, raise ShapeError if not"""
    shape_obt = xs.shape
    if shape_obt != shape_exp:
        raise ShapeError(f"Shape mismatch: expected {shape_exp}, got {shape_obt}")


def prod(x: tuple[int, ...]) -> int:
    """Minor correction to np.prod function in the case where the shape is ().
    prod is used to ensure that the product of a tuple of int outputs an int."""
    return int(np.prod(x))


def get_pre_shape(xs: np.ndarray, exp_shape: tuple[int, ...]) -> tuple[int, ...]:
    r"""For an array of shape (i_1, ..., i_n, e_1, ..., e_m) and shape (e_1, ..., e_m),
    returns the shape (i_1, ..., i_n). This is an helper for vectorized functions.
    If array shape does not end with expected shape, raises ShapeError.
    """
    if exp_shape == ():
        return xs.shape
    n_dim = len(exp_shape)
    tot_shape = xs.shape

    if len(tot_shape) < n_dim:
        raise ShapeError("Shape of input array is not compliant with expected shape")

    if tot_shape[-n_dim:] != exp_shape:
        raise ShapeError("Shape of input array is not compliant with expected shape")

    return tot_shape[:-n_dim]



import signal
from contextlib import contextmanager


@contextmanager
def timeout(duration):
    def timeout_handler(signum, frame):
        raise TimeoutError(f"block timedout after {duration} seconds")

    signal.signal(signal.SIGALRM, timeout_handler)
    signal.alarm(duration)
    try:
        yield
    finally:
        signal.alarm(0)


def timedrun(max_time: int):
    """Decorator to stop function calls if they last too long
    Args:
        max_time: the maximum time for a function call before raising an exception
    Returns:
        A decorator implementing the time out mechanism on the function
    """

    def deco(function):
        def wrap(*args, **kwargs):
            with timeout(max_time):
                return function(*args, **kwargs)

        return wrap

    return deco



from setuptools import setup

long_description = "".join(
    [
        "Miscellaneous functions shared between packages"
    ]
)

setup(
    name="apicutils",
    version="0.0.2",
    author="Antoine Picard-Weibel",
    author_email="apicard.w@gmail.com",
    description="personal misc functions",
    long_description=long_description,
    packages=["apicutils"],
    package_dir={"apicutils": "apicutils"},
    install_requires=[
        "pandas",
        "numpy<=1.26",
        "multiprocess>=0.70",
    ],
    python_requires=">=3.9",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



def blab(silent: bool, *args, **kwargs) -> None:
    """
    Wrap up for print. If silents, does not print, else, prints.
    """
    if not silent:
        print(*args, **kwargs)


def tprint(msg, n: int = 20) -> None:
    """Title print"""
    print()
    print("-" * n, msg, "-" * n)
    print()



from typing import Callable, Optional

import numpy as np
from numpy.typing import ArrayLike

from apicutils.composition import interpretation
from apicutils.parallel import par_eval
from apicutils.shape import prod


def num_der(
    fun: Callable[[ArrayLike], ArrayLike],
    x0: ArrayLike,
    f0: Optional[ArrayLike] = None,  # pylint: disable=W0613
    rel_step: Optional[float] = None,
    parallel: bool = True,
) -> np.ndarray:
    """Return the Jacobian of a function
    If f : shape_x -> shape_y,
    the output is of shape (shape_x, shape_y)

    Arguments:
        fun: the function to derivate
        x0: the point at which to derivate the function
        f0: the value of fun at x0 (is not used since 2 point approximation of the derivative is used)
        parallel: should the evaluations of fun be parallelized
    Output:
        The approximated jacobian of fun at x0 as a np.ndarray of shape (shape_x, shape_y)
    """

    # Work with flat array to facilitate
    x0 = np.array(x0)
    shape_in = x0.shape
    x0 = x0.flatten()

    dim = prod(shape_in)
    loc_fun = interpretation(lambda x: x.reshape(shape_in))(fun)

    if rel_step is None:
        rel_step = float((np.finfo(x0.dtype).eps) ** (1 / 3))

    to_evaluate = np.full((2 * dim, dim), x0)

    delta_x = np.maximum(1.0, x0) * rel_step
    add_matrix = np.diag(delta_x)
    to_evaluate[::2] = to_evaluate[::2] + add_matrix

    to_evaluate[1::2] = to_evaluate[1::2] - add_matrix

    evals = np.array(par_eval(loc_fun, to_evaluate, parallel=parallel))

    der = evals[::2] - evals[1::2]

    for i, d_x in enumerate(delta_x):
        der[i] = der[i] / (2 * d_x)
    shape_out = der[0].shape
    return der.reshape(shape_in + shape_out)



import warnings
from typing import Any, Callable, Union

Input = Any
Output = Any


class SafeCallWarning(Warning):
    """Warning for safe_call context. Enables sending a warning of failure inside a safe_call
    decorated function even if UserWarning is filtered as an exception."""


def safe_call(fun: Callable[[Input], Output]) -> Callable[[Input], Union[None, Output]]:
    """
    Decorator to evaluate a function safely.
    If function call fails, throws a warning and returns None.

    Note:
    Decorated function can still fail IF SafeCallWarning is filtered as an error (which completely
    defeats SafeCallWarning purpose) inside fun.
    """

    def wrapper(*args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except Exception as exc:  # pylint: disable=W0703
            warnings.warn(
                f"Evaluation failed with inputs {args}, {kwargs}: {exc}",
                category=SafeCallWarning,
            )
            return None

    return wrapper
